/*------------------------------------------------------------------------------
* postpos.c : post-processing positioning
*
*          Copyright (C) 2007 by T.TAKASU, All rights reserved.
*
* options : -RCVLOG   use receiver log file
*
* version : $Revision: 1.1 $ $Date: 2008/07/17 21:48:06 $
* history : 2007/05/08  1.0 new
*           2008/06/16  1.1 support binary inputs
*-----------------------------------------------------------------------------*/
#include "rtklib.h"

static const char rcsid[]="$Id: postpos.c,v 1.1 2008/07/17 21:48:06 ttaka Exp $";

static const char prog[]="rtklib ver.2.1.0";

/* types ---------------------------------------------------------------------*/

typedef struct {        /* option parameter type */
    char *name;         /* parameter name */
    int type;           /* parameter type (0:int,1:double,2:string,3:time) */
    void *prm;          /* parameter pointer */
} prm_t;

typedef struct {        /* analysis results struct */
    gtime_t t;          /* time */
    double rr[3],Qr[9]; /* float, dgps or single solution */
    double rf[3],Qf[9]; /* fixed solution */
    double rb[3];       /* base station position */
    int stat,ns;        /* status, number of satellite */
} result_t;

/* application defined functions ---------------------------------------------*/
extern int showmsg(char *format,...);   /* show message and check abort */
extern void settspan(gtime_t ts, gtime_t te); /* set span to show progress */
extern void settime(gtime_t time);      /* set time to show progress */

/* constants/global variables ------------------------------------------------*/
#define ELMINR      15.0    /* elevation mask for reference station (deg) */
#define COMMENTH    "%"     /* comment line indicator */
#define MAXSTD      1.0     /* max standard deviation of fix solution (m) */

static prcopt_t prcopt;     /* processing option */
static solopt_t solopt;     /* solution option */
static gtime_t tstart={0};  /* start time */
static gtime_t tend={0};    /* end time */
static double tint=0.0;     /* time interval (sec) */
static int dirs=0;          /* solutions (0:forward,1:backward,2:combined) */

static FILE *fp;            /* output file pointer */
static obs_t obss;          /* observation data */
static nav_t navs;          /* navigation messages */
static sbs_t sbss;          /* sbas messages */

static int nepoch=0;        /* number of observation epochs */
static int iobsu=0;         /* current rover observation data record index */
static int iobsr=0;         /* current reference observation data record index */
static int isbs=0;          /* current sbas message index */
static int revs=0;          /* analysis direction (0:forward,1:backward) */
static int aborts=0;        /* abort status */
static double tspan=1.0;    /* time span (sec) */
static result_t *resf;      /* forward analysis results */
static result_t *resb;      /* backward analysis results */
static int iresf=0;         /* current forward analysis result index */
static int iresb=0;         /* current backward analysis result index */

/* deg to deg/min/sec --------------------------------------------------------*/
static void todms(double deg, double *dms)
{
    double sgn=deg<0.0?-1.0:1.0;
    
    deg=fabs(deg);
    dms[0]=floor(deg); deg=(deg-dms[0])*60.0;
    dms[1]=floor(deg); dms[2]=(deg-dms[1])*60.0;
    dms[0]*=sgn;
}
/* compute base station position as average of single position ---------------*/
static int refpos(double *rb)
{
    int i,j,k,n;
    double rr[3]={0.0},Q[9],dtr,azel[MAXOBS*NFREQ];
    
    trace(3,"refpos: nobs=%d\n",obss.n);
    
    for (i=n=0;i<obss.n;i=j) {
        for (;i<obss.n;i++) if (obss.data[i].rcv==2) break;
        for (j=i;j<obss.n;j++) {
            if (timediff(obss.data[j].time,obss.data[i].time)>0.0) break;
        }
        if (pntpos(obss.data+i,j-i,&navs,ELMINR*D2R,0.0,rr,Q,&dtr,azel)<0) continue;
        for (k=0;k<3;k++) rb[k]+=rr[k]; n++;
    }
    if (n<=0) {
        trace(3,"refpos: no base position\n");
        return -1;
    }
    for (k=0;k<3;k++) rb[k]/=n;
    return 0;
}
/* output reference position -------------------------------------------------*/
static void outrpos(FILE *fp, const double *r, int posf, int degf)
{
    double pos[3],dms1[3],dms2[3];
    char *sep=solopt.sep;
    
    trace(3,"outrpos: r=%.3f %.3f %.3f\n",r[0],r[1],r[2]);
    
    if (posf==0) {
        ecef2pos(r,pos);
        if (degf) {
            todms(pos[0]*R2D,dms1); todms(pos[1]*R2D,dms2);
            fprintf(fp,"%3.0f%s%02.0f%s%08.5f%s%4.0f%s%02.0f%s%08.5f%s%10.4f",
                    dms1[0],sep,dms1[1],sep,dms1[2],sep,dms2[0],sep,dms2[1],
                    sep,dms2[2],sep,pos[2]);
        }
        else {
            fprintf(fp,"%13.9f%s%14.9f%s%10.4f",pos[0]*R2D,sep,pos[1]*R2D,
                    sep,pos[2]);
        }
    }
    else if (posf==1) {
        fprintf(fp,"%14.4f%s%14.4f%s%14.4f",r[0],sep,r[1],sep,r[2]);
    }
}
/* output header -------------------------------------------------------------*/
static void outhead(char **file, int n)
{
    int i,j,w1,w2;
    char s1[32],s2[32];
    char *s3[]={"single","dgps","kinematic","static","moving-base","fixed"};
    char *s4[]={"L1","L1+L2","L1+L2+L5"},*s5[]={"forward","backward","combined"};
    char *s6[]={"continuous","instantaneous"};
    char *s7[]={"GPST","UTC","JST"};
    double t1,t2;
    gtime_t ts,te;
    
    trace(3,"outhead: file=%s n=%d\n",file[0],n);
    
    if (solopt.outhead) {
        fprintf(fp,"%s program   : %s\n",COMMENTH,prog);
        for (i=0;i<n;i++) {
            fprintf(fp,"%s inp file  : %s\n",COMMENTH,file[i]);
        }
        for (i=0;i<obss.n;i++)    if (obss.data[i].rcv==1) break;
        for (j=obss.n-1;j>=0;j--) if (obss.data[j].rcv==1) break;
        if (j<i) {fprintf(fp,"\n%s no rover obs data\n",COMMENTH); return;}
        ts=obss.data[i].time;
        te=obss.data[j].time;
        t1=time2gpst(ts,&w1);
        t2=time2gpst(te,&w2);
        if (solopt.times>=1) ts=gpst2utc(ts);
        if (solopt.times>=1) te=gpst2utc(te);
        if (solopt.times==2) ts=timeadd(ts,9*3600.0);
        if (solopt.times==2) te=timeadd(te,9*3600.0);
        time2str(ts,s1,1);
        time2str(te,s2,1);
        fprintf(fp,"%s obs start : %s %s (week%04d %8.1fs)\n",COMMENTH,s1,s7[solopt.times],w1,t1);
        fprintf(fp,"%s obs end   : %s %s (week%04d %8.1fs)\n",COMMENTH,s2,s7[solopt.times],w2,t2);
    }
    if (solopt.outopt) {
        fprintf(fp,"%s pos mode  : %s\n",COMMENTH,s3[prcopt.mode]);
        if (prcopt.mode>=1) {
            fprintf(fp,"%s freqs     : %s\n",COMMENTH,s4[prcopt.nf-1]);
            fprintf(fp,"%s solution  : %s\n",COMMENTH,s5[dirs]);
        }
        fprintf(fp,"%s elev mask : %.1f deg\n",COMMENTH,prcopt.elmin*R2D);
        fprintf(fp,"%s snr mask  : %.1f dBHz\n",COMMENTH,prcopt.snrmin);
        if (1<=prcopt.mode) {
            fprintf(fp,"%s ionos est : %s\n",COMMENTH,prcopt.estiono?"on":"off");
            fprintf(fp,"%s tropo est : %s\n",COMMENTH,prcopt.esttrop?"on":"off");
        }
        if (2<=prcopt.mode&&prcopt.mode<=4) {
            fprintf(fp,"%s amb res   : %s\n",COMMENTH,
                    prcopt.thresar<=0.0?"off":s6[prcopt.modear]);
            if (prcopt.thresar>0.0) {
                fprintf(fp,"%s val thres : %.1f\n",COMMENTH,prcopt.thresar);
            }
        }
    }
    if (1<=prcopt.mode&&prcopt.mode<=3) {
        fprintf(fp,"%s ref pos   :",COMMENTH);
        outrpos(fp,prcopt.rb,solopt.posf,solopt.degf);
        fprintf(fp,"\n");
    }
    if (solopt.outhead||solopt.outopt) fprintf(fp,"%s\n",COMMENTH);
    outsolhead(fp);
}
/* search next observation data index ----------------------------------------*/
static int nextobsf(const obs_t *obs, int *i, int rcv)
{
    double tt;
    int n;
    
    for (;*i<obs->n;(*i)++) if (obs->data[*i].rcv==rcv) break;
    for (n=0;*i+n<obs->n;n++) {
        tt=timediff(obs->data[*i+n].time,obs->data[*i].time);
        if (obs->data[*i+n].rcv!=rcv||tt>DTTOL) break;
    }
    return n;
}
static int nextobsb(const obs_t *obs, int *i, int rcv)
{
    double tt;
    int n;
    
    for (;*i>=0;(*i)--) if (obs->data[*i].rcv==rcv) break;
    for (n=0;*i-n>=0;n++) {
        tt=timediff(obs->data[*i-n].time,obs->data[*i].time);
        if (obs->data[*i-n].rcv!=rcv||tt<-DTTOL) break;
    }
    return n;
}
/* input observation data, navigation messages and sbas messages -------------*/
static int infunc(obsd_t **obs, nav_t **nav, sbs_t **sbs)
{
    static obsd_t obse[MAXOBS];
    static sbs_t sbse;
    int i,nu,nr,n=0;
    char s[64];
    
    trace(3,"infunc: revs=%d iobsu=%d iobsr=%d\n",revs,iobsu,iobsr);
    
    if (0<=iobsu&&iobsu<obss.n&&tspan>0.0) {
        settime(obss.data[iobsu].time);
        time2str(obss.data[iobsu].time,s,0);
        if (showmsg("processing : %s",s)) {
            aborts=1; showmsg("aborted"); return 0;
        }
    }
    if (!revs) { /* forward */
        if ((nu=nextobsf(&obss,&iobsu,1))<=0) return 0;
        if (prcopt.intpref) {
            for (;(nr=nextobsf(&obss,&iobsr,2))>0;iobsr+=nr)
                if (timediff(obss.data[iobsr].time,obss.data[iobsu].time)>-DTTOL) break;
        }
        else {
            for (i=iobsr;(nr=nextobsf(&obss,&i,2))>0;iobsr=i,i+=nr)
                if (timediff(obss.data[i].time,obss.data[iobsu].time)>DTTOL) break;
        }
        nr=nextobsf(&obss,&iobsr,2);
        for (i=0;i<nu;i++) obse[n++]=obss.data[iobsu+i];
        for (i=0;i<nr;i++) obse[n++]=obss.data[iobsr+i];
        iobsu+=nu;
        if (sbs&&n>0) {
            for (i=isbs;i<sbss.n;i++) {
                if (timediff(gpst2time(sbss.msgs[i].week,sbss.msgs[i].tow),
                             obse[0].time)>DTTOL) break;
            }
            sbse.n=i-isbs;
            sbse.msgs=sbss.msgs+isbs;
            isbs=i;
        }
        else sbse.n=0;
    }
    else { /* backward */
        if ((nu=nextobsb(&obss,&iobsu,1))<=0) return 0;
        if (prcopt.intpref) {
            for (;(nr=nextobsb(&obss,&iobsr,2))>0;iobsr-=nr)
                if (timediff(obss.data[iobsr].time,obss.data[iobsu].time)<DTTOL) break;
        }
        else {
            for (i=iobsr;(nr=nextobsb(&obss,&i,2))>0;iobsr=i,i-=nr)
                if (timediff(obss.data[i].time,obss.data[iobsu].time)<-DTTOL) break;
        }
        nr=nextobsb(&obss,&iobsr,2);
        for (i=0;i<nu;i++) obse[n++]=obss.data[iobsu-nu+1+i];
        for (i=0;i<nr;i++) obse[n++]=obss.data[iobsr-nr+1+i];
        iobsu-=nu;
        if (sbs&&n>0) {
            for (i=isbs;i>=0;i--) {
                if (timediff(gpst2time(sbss.msgs[i].week,sbss.msgs[i].tow),
                             obse[0].time)<-DTTOL) break;
            }
            sbse.n=isbs-i;
            sbse.msgs=sbss.msgs+i+1;
            isbs=i;
        }
        else sbse.n=0;
    }
    *obs=obse;
    *nav=&navs;
    if (sbs) *sbs=&sbse;
    return n;
}
/* output callback -----------------------------------------------------------*/
static void outfunc(gtime_t time, sol_t sol, sol_t solf, const double *rb,
                    int stat, int ns)
{
    sol_t *p=stat==1?&solf:&sol;
    double rr[3],Qr[9];
    int i,j;
    
    trace(3,"outfunc: stat=%d int=%d\n",stat,ns);
    
    for (i=0;i<3;i++) {
        rr[i]=p->rr[i]; for (j=0;j<3;j++) Qr[i+j*3]=p->Qr[i+j*p->n];
    }
    outsol(fp,time,rr,Qr,rb,stat,ns);
}
/* save result callback ------------------------------------------------------*/
static void savefunc(gtime_t time, sol_t sol, sol_t solf, const double *rb,
                     int stat, int ns)
{
    result_t *p;
    int i,j;
    
    trace(3,"savefunc: stat=%d int=%d\n",stat,ns);
    
    if ((revs?iresb:iresf)>=nepoch) return;
    p=revs?resb+iresb++:resf+iresf++;
    p->t=time; p->stat=stat; p->ns=ns;
    for (i=0;i<3;i++) {
        p->rr[i]=sol.rr[i];  for (j=0;j<3;j++) p->Qr[i+j*3]=sol.Qr[i+j*sol.n];
        p->rf[i]=solf.rr[i]; for (j=0;j<3;j++) p->Qf[i+j*3]=solf.Qr[i+j*solf.n];
        p->rb[i]=rb[i];
    }
}
/* open output file and output header ----------------------------------------*/
static int openfile(const char *outfile, char **infile, int n)
{
    if (!*outfile) fp=stdout;
    else if (!(fp=fopen(outfile,"w"))) {
        showmsg("error : output file open error : %s",outfile);
        return -1;
    }
    /* output header */
    if (solopt.posf!=3) outhead(infile,n);
    return 0;
}
/* combine forward/backward solutions and output results ---------------------*/
static void combres(void)
{
    int i,j,k,stat;
    double tt,dr,rr[3],Qr[9],rb[3];
    
    trace(3,"combres: iresf=%d iresb=%f\n",iresf,iresb);
    
    for (i=0,j=iresb-1;i<iresf&&j>=0;i++,j--) {
        tt=timediff(resf[i].t,resb[j].t);
        if (tt<-DTTOL) {
            if (resf[i].stat==1)
                 outsol(fp,resf[i].t,resf[i].rf,resf[i].Qf,resf[i].rb,resf[i].stat,resf[i].ns);
            else outsol(fp,resf[i].t,resf[i].rr,resf[i].Qr,resf[i].rb,resf[i].stat,resf[i].ns);
            j++;
            continue;
        }
        else if (tt>DTTOL) {
            if (resb[j].stat==1)
                 outsol(fp,resb[j].t,resb[j].rf,resb[j].Qf,resb[j].rb,resb[j].stat,resb[j].ns);
            else outsol(fp,resb[j].t,resb[j].rr,resb[j].Qr,resb[j].rb,resb[j].stat,resb[j].ns);
            i--;
            continue;
        }
        if (resf[i].stat<resb[j].stat) {
            if (resf[i].stat==1)
                 outsol(fp,resf[i].t,resf[i].rf,resf[i].Qf,resf[i].rb,resf[i].stat,resf[i].ns);
            else outsol(fp,resf[i].t,resf[i].rr,resf[i].Qr,resf[i].rb,resf[i].stat,resf[i].ns);
            continue;
        }
        else if (resf[i].stat>resb[j].stat) {
            if (resb[j].stat==1)
                 outsol(fp,resb[j].t,resb[j].rf,resb[j].Qf,resb[j].rb,resb[j].stat,resb[j].ns);
            else outsol(fp,resb[j].t,resb[j].rr,resb[j].Qr,resb[j].rb,resb[j].stat,resb[j].ns);
            continue;
        }
        stat=resf[i].stat;
        for (k=0;k<3;k++) rb[k]=resf[i].rb[k];
        
        /* validation of forward/backward fixed solutions for kinematic mode */
        if (prcopt.mode==2&&stat==1) {
            for (k=0;k<3;k++) {
                dr=resf[i].rf[k]-resb[j].rf[k];
                if (dr*dr>16.0*(resf[i].Qf[k+k*3]+resb[j].Qf[k+k*3]))
                    stat=2; /* degrade to float solution */
            }
        }
        if (stat==1) { /* fixed */
            if (smoother(resf[i].rf,resf[i].Qf,resb[j].rf,resb[j].Qf,3,rr,Qr))
                continue;
        }
        else { /* float,dgps or single */
            if (smoother(resf[i].rr,resf[i].Qr,resb[j].rr,resb[j].Qr,3,rr,Qr))
                continue;
        }
        outsol(fp,timeadd(resf[i].t,-tt/2.0),rr,Qr,rb,stat,resf[i].ns);
    }
}
/* execute positioning -------------------------------------------------------*/
static int execpos(char **infile, int n, char *outfile, int *informat)
{
    int i,j;
    
    trace(3,"execpos: infile=%s n=%d outfile=%s\n",infile[0],n,outfile);
    
    /* read input files */
#ifdef RCVLOG
    nepoch=readlogs(infile,informat,n,tstart,tend,tint,0,&obss,&navs,&sbss);
#else
    nepoch=readrnxt(infile,n,tstart,tend,tint,&obss,&navs);
#endif
    if (nepoch==0) {
        showmsg("error : no observation data");
        trace(1,"execpos: no observation data\n");
        return -3;
    }
    else if (nepoch<0) {
        showmsg("error : observation data/nav message read error");
        trace(1,"execpos: observation data /nav message read error\n");
        return -4;
    }
    if (navs.n<=0) {
        showmsg("error : no navigation message");
        trace(1,"execpos: no navigation message\n");
        return -5;
    }
    /* set time span */
    for (i=0;i<obss.n;i++)    if (obss.data[i].rcv==1) break;
    for (j=obss.n-1;j>=0;j--) if (obss.data[j].rcv==1) break;
    if (i<j) settspan(obss.data[i].time,obss.data[j].time);
    
    /* code smoothing */
    if (prcopt.codesmooth>0) {
        csmooth(&obss,prcopt.codesmooth);
    }
    /* compute reference station position */
    if (1<=prcopt.mode&&prcopt.mode<=3&&norm(prcopt.rb,3)<=0.0&&refpos(prcopt.rb)) {
        showmsg("error : base position computation");
        trace(1,"execpos: base position computation\n");
        obss.n=navs.n=sbss.n=0;
        free(obss.data); free(navs.eph); free(sbss.msgs);
        obss.data=NULL; navs.eph=NULL; sbss.msgs=NULL;
        return -6;
    }
    for (i=0,tstart.time=0;i<obss.n;i++) {
        if (obss.data[i].rcv!=1) continue;
        if (tstart.time==0) tstart=obss.data[i].time;
        tspan=timediff(obss.data[i].time,tstart);
    }
    showmsg("processing :");
    
    iobsu=iobsr=isbs=revs=aborts=0;
    if (dirs<2) { /* forward/backward solution */
        if (openfile(outfile,infile,n)) return -6;
        if (dirs) {
            revs=1; iobsu=iobsr=obss.n-1; isbs=sbss.n-1;
        }
        rtkpos(infunc,outfunc,&prcopt);
        fclose(fp);
    }
    else { /* combined solution */
        resf=(result_t *)malloc(sizeof(result_t)*nepoch);
        resb=(result_t *)malloc(sizeof(result_t)*nepoch);
        if (resf&&resb) {
            iresf=iresb=0;
            rtkpos(infunc,savefunc,&prcopt);
            revs=1; iobsu=iobsr=obss.n-1; isbs=sbss.n-1;
            rtkpos(infunc,savefunc,&prcopt);
            
            if (!aborts&&!openfile(outfile,infile,n)) {
                /* combine forward/backward solutions and output results */
                combres();
                fclose(fp);
            }
        }
        else showmsg("error : memory allocation error");
        free(resf); free(resb);
    }
    free(obss.data); free(navs.eph); free(sbss.msgs);
    obss.data=NULL; navs.eph=NULL; sbss.msgs=NULL;
    obss.n=navs.n=sbss.n=0;
    
    return aborts?1:0;
}
/* post positioning -------------------------------------------------------------
* post-processing positioning
* args   : gtime_t ts,te    I   processing start/end time
*          double tint      I   processing interval (s) (0:all)
*          int    dirs      I   proc direction (0:forward,1:backward,2:combined)
*          prcopt_t *prcopt I   processing options
*          solopt_t *solopt I   solution options
*          char   *infile[] I   input files
*          int    nfile     I   number of input files
*          char   *outfile  I   output file ("":stdout)
*          int    *informat I   input file formats
* return : status (0:ok,0>:error,1:aborted)
*-----------------------------------------------------------------------------*/
extern int postpos(gtime_t ts_, gtime_t te_, double tint_, int dirs_,
                   const prcopt_t *prcopt_, const solopt_t *solopt_,
                   char **infile, int nfile, char *outfile, int *informat)
{
    tstart=ts_; tend=te_; tint=tint_; dirs =dirs_; prcopt=*prcopt_;
    solopt=*solopt_; setsolopt(solopt_);
    return execpos(infile,nfile,outfile,informat);
}
/* processing option buffer --------------------------------------------------*/
static prcopt_t opts1;
static solopt_t opts2;
static const prm_t prm1[]={
    {"positioning mode",               0,(void *)&opts1.mode},
    {"frequencies",                    0,(void *)&opts1.nf},
    {"elevation mask",                 1,(void *)&opts1.elmin},
    {"snr mask",                       1,(void *)&opts1.snrmin},
    {"ionosphere estimation",          0,(void *)&opts1.estiono},
    {"troposphere estimation",         0,(void *)&opts1.esttrop},
    {"integer ambiguity resolution",   0,(void *)&opts1.modear},
    {"validation threshold to fix ambiguity",1,(void *)&opts1.thresar},
    {"minimum lock count to fix ambiguity",0,(void *)&opts1.minlock},
    {"minimum elevation to fix ambiguity",1,(void *)&opts1.elmaskar},
    {"outage count to reset ambiguity",0,(void *)&opts1.maxout},
    {"slip threshold of gf difference",1,(void *)&opts1.thresslip},
    {"max age of differential",        1,(void *)&opts1.maxtdiff},
    {"reject threshold of innovation", 1,(void *)&opts1.maxinno},
    {"number of filter iterationn",    0,(void *)&opts1.niter},
    {"measurement error 1",            1,(void *)&opts1.err[0]},
    {"measurement error 2",            1,(void *)&opts1.err[1]},
    {"measurement error 3",            1,(void *)&opts1.err[2]},
    {"measurement error 4",            1,(void *)&opts1.err[3]},
    {"process noise 1",                1,(void *)&opts1.prn[0]},
    {"process noise 2",                1,(void *)&opts1.prn[1]},
    {"process noise 3",                1,(void *)&opts1.prn[2]},
    {"satellite clock stability",      1,(void *)&opts1.sclkstab},
    {"base station position x",        1,(void *)&opts1.rb[0]},
    {"base station position y",        1,(void *)&opts1.rb[1]},
    {"base station position z",        1,(void *)&opts1.rb[2]},
    {"time interpolation of base station obs",0,(void *)&opts1.intpref},
    {"sbas correction",                0,(void *)&opts1.sbascorr}
};
static const prm_t prm2[]={
    {"solution format",                0,(void *)&opts2.posf},
    {"output header",                  0,(void *)&opts2.outhead},
    {"output processing option",       0,(void *)&opts2.outopt},
    {"time system",                    0,(void *)&opts2.times},
    {"time format",                    0,(void *)&opts2.timef},
    {"time digits under decimal point",0,(void *)&opts2.timeu},
    {"latitude/longitude format",      0,(void *)&opts2.degf},
    {"field separator",                2,(void *)&opts2.sep},
    {"datum",                          0,(void *)&opts2.datum},
    {"height",                         0,(void *)&opts2.height},
    {"geoid model",                    0,(void *)&opts2.geoid}
};
/* read parameters -----------------------------------------------------------*/
static void readprm(FILE *fp, const prm_t *prm, int n)
{
    int i;
    char buff[256],*p,*q;
    
    while (fgets(buff,sizeof(buff),fp)) {
        if (strncmp(buff,"%",1)) continue; /* comment line */
        if ((p=strchr(buff,':'))) *p='\0'; else continue;
        for (i=0;i<n;i++) {
            if (strstr(buff,prm[i].name)!=buff) continue;
            if ((q=strchr(p+1,'%'))) *q='\0';
            switch (prm[i].type) {
            case 0: sscanf(p+1,"%d", (int *)prm[i].prm); break;
            case 1: sscanf(p+1,"%lf",(double *)prm[i].prm); break;
            case 2: strcpy((char *)prm[i].prm,p+1); break;
            }
        }
    }
}
/* write parameters ----------------------------------------------------------*/
static void writeprm(FILE *fp, const prm_t *prm, int n)
{
    int i;
    
    for (i=0;i<n;i++) {
        fprintf(fp,"%-32s:",prm[i].name);
        switch (prm[i].type) {
        case 0: fprintf(fp,"%d",*((int *)prm[i].prm)); break;
        case 1: fprintf(fp,"%f",*((double *)prm[i].prm)); break;
        case 2: fprintf(fp,"%s",(char *)prm[i].prm); break;
        }
    }
}
/* read processing/solution options --------------------------------------------
* read processing/solution options from file
* args   : char   *file     I   options file
*          prcopt_t  *prcopt O  processing options
*          solopt_t  *solopt O  solution options
* return : status (0:ok,0>:error)
*-----------------------------------------------------------------------------*/
extern int readopts(const char *file, prcopt_t *prcopt, solopt_t *solopt)
{
    FILE *fp;
    
    opts1=prcopt_default;
    opts2=solopt_default;
    if (!(fp=fopen(file,"r"))) {
        showmsg("options file read error : %s\n",file);
        return -1;
    }
    readprm(fp,prm1,sizeof(prm1)/sizeof(*prm1));
    rewind(fp);
    readprm(fp,prm2,sizeof(prm2)/sizeof(*prm2));
    *prcopt=opts1;
    *solopt=opts2;
    fclose(fp);
    return 0;
}
/* save processing options -----------------------------------------------------
* save processing/solution options to file
* args   : char   *file     I   options file
*          prcopt_t  *prcopt I  processing options
*          solopt_t  *solopt I  solution options
* return : status (0:ok,0>:error)
*-----------------------------------------------------------------------------*/
extern int writeopts(const char *file, const prcopt_t *prcopt,
                     const solopt_t *solopt)
{
    FILE *fp;
    
    if (!(fp=fopen(file,"w"))) {
        showmsg("options file save error : %s\n",file);
        return -1;
    }
    opts1=*prcopt;
    opts2=*solopt;
    writeprm(fp,prm1,sizeof(prm1)/sizeof(*prm1));
    writeprm(fp,prm2,sizeof(prm2)/sizeof(*prm2));
    fclose(fp);
    return 0;
}
